/* bf.c
******************************************************************************

   Yet another Brainfuck interpreter

   Author: Stephan Beyer
   
   Copyright (C) GPL, 2003, 2004 Stephan Beyer - s-beyer@gmx.net
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h> /* malloc, atoi */
#include <string.h> /* strlen */

#include <slang.h> /* S-Lang library */
#if 0
#include <curses.h> /* GNU NCurses */
#endif

#include "bf.h"
#include "errors.h"

/* global variables:
 * Brainfuck array, pointer and bf options */
char *array;
char *p;
Opt opt;

/* begins interpretion in filehandle fp at <start> end returns where it
 * ended (either at ] or eof - on eof it returns -1)  */
long Interprete (FILE *fp, long start)
{
	int c;
	int do_nothing = 0;

	DoAndErr(fseek(fp, start, SEEK_SET), fseek, errFILE);

	while ((c = fgetc(fp)) != EOF)
	{
		if ((do_nothing) && (c != ']')) {
			if (c == '[')
				do_nothing++;
			continue;
		}

		switch(c)
		{
			case '>':
				if (p+1 > array+opt.cells)
					ErrorMsg(errOUTOFRANGE, "Out of range! You wanted to '>' beyond the last cell. See -c option.");
				++p;
				ShowInput(c);
				break;
			case '<':
				if (p-1 < array)
					ErrorMsg(errOUTOFRANGE, "Out of range! You wanted to '<' below the first cell.\nTo solve, add some '>'s at the beginning, for example.");
				--p;
				ShowInput(c);
				break;
			case '+':
				if ((!opt.wraparound) && (*p == 255))
					ErrorMsg(errWRAPAROUND, "Out of range! You wanted to '+' a 0xFF byte. See -w option.");
				++*p;
				ShowInput(c);
				break;
			case '-':
				if ((!opt.wraparound) && (!*p))
					ErrorMsg(errWRAPAROUND, "Out of range! You wanted to '-' a 0x00 byte. See -w option.");
				--*p;
				ShowInput(c);
				break;
			case '.':
				ShowInput(c);
				putchar(*p);
				fflush(stdout);
				break;
			case ',':
				ShowInput(c);

				if ((opt.inputmode >= 1)
				 && (opt.inputmode <= 4))
				{
					/* We init and reset the terminal
					 * on each input (and not once at 
					 * start and end of program) to
					 * prevent bugs with bf output... */
					
					/* init slang terminal */
					DoAndErr(SLang_init_tty(-1, 0, 1) ==
						-1, SLang_init_tty, errSLANG);
					
					/* get any key */
					if (opt.inputmode <= 2) /* 1 or 2 */
						*p = SLang_getkey();
					else /* 3 or 4 */
					{ /* get key but no escape characters */
						int tmp;
						do {
							tmp = SLkp_getkey();
						} while(tmp > 255);
						*p = (unsigned char)tmp;
					}
					
					if (!(opt.inputmode % 2)) /* 2 and 4 */
					{
						putchar(*p); /* print them */
						fflush(stdout);
					}
					
					/* and reset */
					SLang_reset_tty();
					if (*p == '\r')
						printf("\n");
				}
				else /* opt.inputmode == 0 (default) */
					*p = getchar();
				break;
			case '[':
			{
				long pos = ftell(fp);
				ShowInput(c);

				if (!*p)
				{
					do_nothing++;
					Debug(skip loop)
					break;
				}
				Debug(begin loop);

				while(*p) {
					Debug(loop)
					Interprete(fp, pos);
				}

				Debug(loop ended);

				/* we should still be at the position where
				 * we ought to be, so I won't fseek */
			}	break;
			case ']':
				if (do_nothing)
					do_nothing--;
				else {
					ShowInput(c);
					return ftell(fp);
				}
		}
	}
	return -1;
}

void ResetArray()
{
	int i;
	array = malloc(opt.cells+1);
	DoAndErr(!array, malloc, errMEMORY);
	p = array;

	for (i = 0; i < opt.cells; i++)
		*(p++) = 0;

	p = array;
}

/* display usage information */
void Usage(char *bin)
{
	printf	("Yet another Brainfuck interpreter       version %s\n", VERSION);
	puts	("(C) 2003, 2004, Stephan Beyer, GPL, s-beyer@gmx.net\n");
	puts	("Usage information: ");
	printf	("\t%s [-h] [options] inputfile\n\n", bin);
	puts	("Available options:");
	printf	("\t-c<num>   specify number of cells [%d]\n", opt.cells);
	puts	("\t-d        show some debugging information (stderr)");
	puts	("\t-i        show used input (stderr)");
	puts	("\t-w        disallow decrementing 0 and incrementing 255");
	printf	("\t-,<mode>  set input mode: 0-4 [%d]\n", opt.inputmode);
	puts	("\nSee the bf(1) manpage for more information.");
	puts	("Have fun!");
}

void HandleOptions(int c, char **v)
/* too lazy to use getopt this time ;) */
{
	int i;
	/* setting defaults */
	opt.showinput = 0;
	opt.cells = CELLS;
	opt.inputmode = 4;
	opt.debug = 0;
	opt.wraparound = 1; /* allow by default */
	
	if (c <= 1) /* no filename given */
	{
		Usage(*v);
		ErrorMsg (errOPT, "No input file given.");
	}
	if (!strcmp(*(v+1), "-h")) /* is first argument '-h'? */
	{
		Usage(*v);
		exit(0);
	}
	
	for(i = 1; i < c-1; i++) /* options? */
	{
		if ((**(v+i) == '-')
		 && (strlen(*(v+i)) >= 2))
		{
			switch(*(*(v+i)+1))
			{
				case 'c':
					opt.cells = atoi(*(v+i)+2);
					break;
				case 'd':
					opt.debug = 1;
					break;
				case 'i':
					opt.showinput = 1;
					break;
				case 'w':
					opt.wraparound = 0;
					break;
				case ',':
					opt.inputmode = atoi(*(v+i)+2);
					break;
				default:
					Usage(*v);
					printf("Unable to handle option %s - ",
						*(v+i));
					ErrorMsg(errOPT, "Unknown option.");
			}
		} else {
			Usage(*v);
			printf("Invalid argument %s - ", *(v+i));
			ErrorMsg(errOPT, "Invalid argument(s).");
		}
	}
	opt.filename = *(v+i); /* last option MUST be filename */
}

int main(int argc, char **argv)
{
	FILE *fp;

	/* options */
	HandleOptions(argc, argv);
	/* init slang kp if mode 3 or 4 */
	if ((opt.inputmode == 3)
	 || (opt.inputmode == 4))
	{
		SLtt_get_terminfo();
		DoAndErr(SLkp_init() == -1,
			SLkp_init, errSLANG);
	}

	ResetArray();

	/* file opening banana */
	fp = fopen(opt.filename, "r");
	DoAndErr(!fp, fopen, errFILE); 

	/* reading and interpreting */
	if(Interprete(fp, 0) > -1)
		ErrorMsg(errBRACKET, "Unmatching brackets!");

	/* cleanup */
	fclose(fp);
	free(array);
	return 0;
}
